#ifndef UAE_FUNCATTR_H
#define UAE_FUNCATTR_H

#include "uae/attributes.h"
/*
#ifdef __GNUC__
#define UAE_PRINTF_FORMAT(f, a) __attribute__((format(printf, f, a)))
#else
#define UAE_PRINTF_FORMAT(f, a)
#endif
*/
/*
#ifdef _MSC_VER
#define UAECALL __cdecl
#else
#define UAECALL
#endif
*/

#endif /* UAE_FUNCATTR_H */
