#ifndef FS_ML_OPTIONS_H
#define FS_ML_OPTIONS_H

#define OPTION_CURSOR "cursor"
#define OPTION_LOG_INPUT "log_input"
#define OPTION_MODIFIER_KEY "modifier_key"
#define OPTION_MOUSE_INTEGRATION "mouse_integration"
#define OPTION_RAW_INPUT "raw_input"

/* Deprecated options */

//

#endif /* FS_ML_OPTIONS_H */
