#ifndef FS_EMU_OPTIONS_H
#define FS_EMU_OPTIONS_H

#define OPTION_AUDIO_DRIVER "audio_driver"
#define OPTION_BEZEL "bezel"
#define OPTION_FULL_KEYBOARD "full_keyboard"
#define OPTION_FULLSCREEN "fullscreen"
#define OPTION_FULLSCREEN_MODE "fullscreen_mode"
#define OPTION_NOTIFICATION_DURATION "notification_duration"
#define OPTION_SCREENSHOTS_OUTPUT_DIR "screenshots_output_dir"
#define OPTION_STDOUT "stdout"
#define OPTION_SCALE "scale"
#define OPTION_STRETCH "stretch"
#define OPTION_SUB_TITLE "sub_title"
#define OPTION_TITLE "title"
#define OPTION_VIDEO_DRIVER "video_driver"

#include <fs/ml/options.h>

#endif /* FS_EMU_OPTIONS_H */
