#ifndef FSE_LOG_H
#define FSE_LOG_H

#include <fs/log.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus
}
#endif

#define fse_log fs_log
#define fse_log_string fs_log_string

#endif /* FSE_LOG_H */
