/*
 * E-UAE - The portable Amiga Emulator
 *
 * Processor-specific definitions
 */

#ifndef MACHDEP_MACHDEP_H
#define MACHDEP_MACHDEP_H

#include <stdint.h>

#define MACHDEP_X86
#define MACHDEP_NAME    "x86"

#define HAVE_MACHDEP_TIMER

#endif
