#ifndef UAE_MODULERIPPER_H
#define UAE_MODULERIPPER_H

#include "uae/api.h"

UAE_EXTERN_C int prowizard_search (unsigned char *in_data, int in_size);
extern void moduleripper (void);

#endif /* UAE_MODULERIPPER_H */
